# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from . import common
from .. import rw
from ..enum import enum
from .base import BaseMessage
from .types import Types

ErrorCode = enum(
    'ErrorCode',
    timeout=0x01,
    cancelled=0x02,
    busy=0x03,
    declined=0x04,
    unexpected=0x05,
    bad_request=0x06,
    network_error=0x07,
    unhealthy=0x08,
    fatal=0xff,
)

error_code_rw = rw.number(1)


class ErrorMessage(BaseMessage):
    """Respond to a CALL_REQ with a failure at the protocol level."""
    message_type = Types.ERROR

    __slots__ = BaseMessage.__slots__ + (
        'code',
        'tracing',
        'description',
    )

    ERROR_CODES = {
        ErrorCode.timeout: 'timeout',
        ErrorCode.cancelled: 'cancelled',
        ErrorCode.busy: 'busy',
        ErrorCode.declined: 'declined',
        ErrorCode.unexpected: 'unexpected',
        ErrorCode.bad_request: 'bad request',
        ErrorCode.network_error: 'network error',
        ErrorCode.unhealthy: 'unhealthy error',
        ErrorCode.fatal: 'fatal protocol error'
    }

    def __init__(self, code=None, tracing=None, description=None, id=0):
        super(ErrorMessage, self).__init__(id)
        self.code = code if code else ErrorCode.unexpected
        self.description = description or ''
        self.tracing = tracing or common.Tracing(0, 0, 0, 0)

    def error_name(self):
        """Get a friendly error message."""
        return self.ERROR_CODES.get(self.code)


error_rw = rw.instance(
    ErrorMessage,
    ('code', error_code_rw),                            # code:1
    ('tracing', common.tracing_rw),                     # tracing:24
    ('description', rw.len_prefixed_string(rw.number(2)))   # message~2
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from collections import namedtuple

import tornado
import tornado.gen
from tornado.iostream import StreamClosedError

from tchannel import retry

from ..glossary import DEFAULT_TIMEOUT
from ..messages import ErrorCode
from ..messages.common import ChecksumType
from ..messages.common import FlagsType
from ..messages.common import StreamState
from ..serializer.raw import RawSerializer
from ..zipkin.trace import Trace
from .stream import InMemStream
from .util import get_arg


class Request(object):
    """Represents an incoming request to an endpoint.

    Request class is used to represent the CallRequestMessage at User's level.
    This is going to hide the protocol level message information.
    """

    # TODO deprecate in favor of top-level tchannel.Request
    # TODO decide which elements inside "message" object to expose to user.

    def __init__(
        self,
        id=None,
        flags=FlagsType.none,
        ttl=DEFAULT_TIMEOUT,
        tracing=None,
        service=None,
        headers=None,
        checksum=None,
        argstreams=None,
        serializer=None,
        endpoint=None,
    ):
        self.flags = flags
        self.ttl = ttl
        self.service = service
        self.tracing = tracing or Trace()
        # argstreams is a list of InMemStream/PipeStream objects
        self.argstreams = argstreams or [InMemStream(),
                                         InMemStream(),
                                         InMemStream()]
        self.checksum = checksum or (ChecksumType.crc32c, 0)
        self.id = id
        self.headers = headers or {}
        self.state = StreamState.init
        self.serializer = serializer or RawSerializer()

        self.is_streaming_request = self._is_streaming_request()
        if not self.is_streaming_request:
            self._copy_argstreams = [
                self.argstreams[0].clone(),
                self.argstreams[1].clone(),
                self.argstreams[2].clone(),
            ]

        self.endpoint = endpoint or ""

    def rewind(self, id=None):
        self.id = id
        if not self.is_streaming_request:
            self.argstreams = [
                self._copy_argstreams[0].clone(),
                self._copy_argstreams[1].clone(),
                self._copy_argstreams[2].clone(),
            ]
        self.state = StreamState.init
        self.tracing = Trace()

    @property
    def arg_scheme(self):
        return self.headers.get('as', None)

    def set_exception(self, exception):
        for stream in self.argstreams:
            stream.set_exception(exception)
            stream.close()

    def close_argstreams(self, force=False):
        for stream in self.argstreams:
            if stream.auto_close or force:
                stream.close()

    @tornado.gen.coroutine
    def get_header(self):
        """Get the header value from the request.

        :return: a future contains the deserialized value of header
        """
        raw_header = yield get_arg(self, 1)
        if not self.serializer:
            raise tornado.gen.Return(raw_header)
        else:
            header = self.serializer.deserialize_header(raw_header)
            raise tornado.gen.Return(header)

    @tornado.gen.coroutine
    def get_body(self):
        """Get the body value from the resquest.

        :return: a future contains the deserialized value of body
        """

        raw_body = yield get_arg(self, 2)
        if not self.serializer:
            raise tornado.gen.Return(raw_body)
        else:
            body = self.serializer.deserialize_body(raw_body)
            raise tornado.gen.Return(body)

    def get_header_s(self):
        """Get the raw stream of header.

        :return: the argstream of header
        """
        return self.argstreams[1]

    def get_body_s(self):
        """Get the raw stream of body.

        :return: the argstream of body
        """
        return self.argstreams[2]

    def _is_streaming_request(self):
        """check request is stream request or not"""
        arg2 = self.argstreams[1]
        arg3 = self.argstreams[2]
        return not (isinstance(arg2, InMemStream) and
                    isinstance(arg3, InMemStream) and
                    ((arg2.auto_close and arg3.auto_close) or (
                        arg2.state == StreamState.completed and
                        arg3.state == StreamState.completed)))

    def should_retry_on_error(self, error):
        """rules for retry

        :param error:
            ProtocolException that returns from Server
        """

        if self.is_streaming_request:
            # not retry for streaming request
            return False

        retry_flag = self.headers.get('re', retry.DEFAULT)

        if retry_flag == retry.NEVER:
            return False

        if isinstance(error, StreamClosedError):
            return True

        if error.code in [ErrorCode.bad_request, ErrorCode.cancelled,
                          ErrorCode.unhealthy]:
            return False
        elif error.code in [ErrorCode.busy, ErrorCode.declined]:
            return True
        elif error.code is ErrorCode.timeout:
            return retry_flag is not retry.CONNECTION_ERROR
        elif error.code in [ErrorCode.network_error,
                            ErrorCode.fatal,
                            ErrorCode.unexpected]:
            return retry_flag is not retry.TIMEOUT
        else:
            return False


class TransportMetadata(
    namedtuple('_Metadata', 'flags ttl service id headers')
):
    """A read-only representation of the metadata contained in the Request."""

    @classmethod
    def from_request(cls, request):
        return cls(
            flags=request.flags,
            ttl=request.ttl,
            service=request.service,
            id=request.id,
            headers=request.headers,
        )



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import inspect
import logging
import os
import socket
import sys
from functools import partial

import tornado.gen
import tornado.ioloop
import tornado.iostream
import tornado.tcpserver
from tornado.netutil import bind_sockets

from . import hyperbahn
from ..context import RequestContextProvider
from ..deprecate import deprecate
from ..enum import enum
from ..errors import AlreadyListeningError
from ..event import EventEmitter
from ..event import EventRegistrar
from ..glossary import (
    TCHANNEL_LANGUAGE,
    TCHANNEL_LANGUAGE_VERSION,
    TCHANNEL_VERSION,
)
from ..net import local_ip
from ..schemes import DEFAULT_NAMES
from ..schemes import JSON
from ..serializer.json import JsonSerializer
from ..serializer.raw import RawSerializer
from .connection import StreamConnection
from .connection import INCOMING
from .dispatch import RequestDispatcher
from .peer import PeerGroup

log = logging.getLogger('tchannel')


State = enum(
    'State',
    ready=0,
    closing=1,
    closed=2,
)


class TChannel(object):
    """Manages inbound and outbound connections to various hosts."""

    # TODO deprecate in favor of top-level tchannel.TChannel

    FALLBACK = RequestDispatcher.FALLBACK

    def __init__(self, name, hostport=None, process_name=None,
                 known_peers=None, trace=False, dispatcher=None,
                 reuse_port=False, context_provider_fn=None,
                 _from_new_api=False):
        """Build or re-use a TChannel.

        :param name:
            Name is used to identify client or service itself.

        :param hostport:
            The host-port at which the service behind this TChannel is
            reachable. The port specified in the ``hostport`` is what the
            server will listen on. If unspecified, the system will attempt to
            determine the local network IP for this host and use an
            OS-assigned port.

        :param process_name:
            Name of this process. This is used for logging only. If
            unspecified, this will default to ``$processName[$processId]``.

        :param known_peers:
            A list of host-ports at which already known peers can be reached.
            Defaults to an empty list.

        :param trace:
            Flag to turn on/off zipkin trace. It can be a bool variable or
            a function that return true or false.

        :param context_provider_fn:
            A getter function to retrieve an instance of
            context.RequestContextProvider used to manage
            thread-local request context.
        """

        self._state = State.ready
        if context_provider_fn:
            self.context_provider_fn = context_provider_fn
        else:
            context_provider = RequestContextProvider()
            self.context_provider_fn = lambda: context_provider

        if not dispatcher:
            self._handler = RequestDispatcher(
                context_provider_fn=lambda: self.context_provider_fn())
        else:
            self._handler = dispatcher

        self.peers = PeerGroup(self)

        self._port = 0
        self._host = None
        if hostport:
            self._host, port = hostport.rsplit(':', 1)
            self._port = int(port)

        if not self._host:
            # TChannel(":4040") => determine IP automatically but use port
            # 4040
            self._host = local_ip()

        self.process_name = process_name or "%s[%s]" % (
            sys.argv[0], os.getpid()
        )

        self.name = name
        self._trace = trace

        # register event hooks
        self.event_emitter = EventEmitter()
        self.hooks = EventRegistrar(self.event_emitter)

        if known_peers:
            for peer_hostport in known_peers:
                self.peers.add(peer_hostport)

        # server created from calling listen()
        self._server = None

        # allow SO_REUSEPORT
        self._reuse_port = reuse_port

        # warn if customers are still using this old and soon to be deleted api
        if _from_new_api is False:
            deprecate(
                "tchannel.tornado.TChannel is deprecated and will be removed" +
                " in a a future version - please switch usage to " +
                "tchannel.TChannel object. Thank you."
            )

    @property
    def trace(self):
        if callable(self._trace):
            return self._trace()
        else:
            return self._trace

    def advertise(
        self,
        routers=None,
        name=None,
        timeout=None,
        router_file=None
    ):
        """Make a service available on the Hyperbahn routing mesh.

        This will make contact with a Hyperbahn host from a list of known
        Hyperbahn routers. Additional Hyperbahn connections will be established
        once contact has been made with the network.

        :param router:
            A seed list of addresses of Hyperbahn routers, e.g.,
            ``["127.0.0.1:23000"]``.

        :param name:
            The identity of this service on the Hyperbahn.

            This is usually unnecessary, as it defaults to the name given when
            initializing the :py:class:`TChannel` (which is used as your
            identity as a caller).

        :returns:
            A future that resolves to the remote server's response after
            the first advertise finishes.

            Advertisement will continue to happen periodically.
        """
        name = name or self.name

        if not self.is_listening():
            self.listen()

        return hyperbahn.advertise(self, name, routers, timeout, router_file)

    @property
    def closed(self):
        return self._state == State.closed

    def close(self):
        if self._state in [State.closed, State.closing]:
            return

        self._state = State.closing
        try:
            self.peers.clear()
            if self._server:
                self._server.stop()
        finally:
            self._state = State.closed

    @property
    def host(self):
        return self._host

    @property
    def hostport(self):
        return "%s:%d" % (self._host, self._port)

    @property
    def port(self):
        return self._port

    def request(self,
                hostport=None,
                service=None,
                arg_scheme=None,
                retry=None,
                **kwargs):
        """Initiate a new request through this TChannel.

        :param hostport:
            Host to which the request will be made. If unspecified, a random
            known peer will be picked. This is not necessary if using
            Hyperbahn.

        :param service:
            The name of a service available on Hyperbahn. Defaults to an empty
            string.

        :param arg_scheme:
            Determines the serialization scheme for the request. One of 'raw',
            'json', or 'thrift'. Defaults to 'raw'.

        :param rety:
            One of 'n' (never retry), 'c' (retry on connection errors), 't'
            (retry on timeout), 'ct' (retry on connection errors and timeouts).

            Defaults to 'c'.
        """
        # TODO disallow certain parameters or don't propagate them backwards.
        # For example, blacklist and rank threshold aren't really
        # user-configurable right now.
        return self.peers.request(hostport=hostport,
                                  service=service,
                                  arg_scheme=arg_scheme,
                                  retry=retry,
                                  **kwargs)

    def listen(self, port=None):
        """Start listening for incoming connections.

        A request handler must have already been specified with
        ``TChannel.host``.

        :param port:
            An explicit port to listen on. This is unnecessary when advertising
            on Hyperbahn.

        :returns:
            Returns immediately.

        :raises AlreadyListeningError:
            If listen was already called.
        """

        if self.is_listening():
            raise AlreadyListeningError(
                "listen has already been called"
            )

        if port:
            assert not self._port, "Port has already been set."
            self._port = int(port)

        assert self._handler, "Call .host with a RequestHandler first"
        server = TChannelServer(self)

        bind_sockets_kwargs = {
            'port': self._port,
            # ipv6 causes random address already in use (socket.error w errno
            # == 98) when getaddrinfo() returns multiple values
            # @see https://github.com/uber/tchannel-python/issues/256
            'family': socket.AF_INET,
        }
        if self._reuse_port is True:
            # allow multiple processes to share the same port,
            # this is really useful in a world where services launch N
            # processes per container/os-space, where N is
            # the amount of cpus for example
            bind_sockets_kwargs['reuse_port'] = True

        sockets = bind_sockets(**bind_sockets_kwargs)
        assert sockets, "No sockets bound for port %d" % self._port

        # If port was 0, the OS probably assigned something better.
        self._port = sockets[0].getsockname()[1]

        server.add_sockets(sockets)

        # assign server so we don't listen twice
        self._server = server

    def is_listening(self):

        if self._server:
            return True

        return False

    def receive_call(self, message, connection):

        if not self._handler:
            log.warn(
                "Received %s but a handler has not been defined.", str(message)
            )
            return
        return self._handler.handle(message, connection)

    def _register_simple(self, endpoint, scheme, f):
        """Register a simple endpoint with this TChannel.

        :param endpoint:
            Name of the endpoint being registered.
        :param scheme:
            Name of the arg scheme under which the endpoint will be
            registered.
        :param f:
            Callable handler for the endpoint.
        """
        assert scheme in DEFAULT_NAMES, ("Unsupported arg scheme %s" % scheme)
        if scheme == JSON:
            req_serializer = JsonSerializer()
            resp_serializer = JsonSerializer()
        else:
            req_serializer = RawSerializer()
            resp_serializer = RawSerializer()
        self._handler.register(endpoint, f, req_serializer, resp_serializer)
        return f

    def _register_thrift(self, service_module, handler, **kwargs):
        """Register a Thrift endpoint on this TChannel.

        :param service_module:
            Reference to the Thrift-generated module for the service being
            registered.
        :param handler:
            Handler for the endpoint
        :param method:
            Name of the Thrift method being registered. If omitted, ``f``'s
            name is assumed to be the method name.
        :param service:
            Name of the Thrift service. By default this is determined
            automatically from the module name.
        """
        import tchannel.thrift as thrift
        # Imported inside the function so that we don't have a hard dependency
        # on the Thrift library. This function is usable only if the Thrift
        # library is installed.
        thrift.register(self._handler, service_module, handler, **kwargs)
        return handler

    def register(self, endpoint, scheme=None, handler=None, **kwargs):
        """Register a handler with this TChannel.

        This may be used as a decorator:

        .. code-block:: python

            app = TChannel(name='bar')

            @app.register("hello", "json")
            def hello_handler(request, response):
                params = yield request.get_body()

        Or as a function:

        .. code-block:: python

            # Here we have a Thrift handler for `Foo::hello`
            app.register(Foo, "hello", hello_thrift_handler)

        :param endpoint:
            Name of the endpoint being registered. This should be a reference
            to the Thrift-generated module if this is a Thrift endpoint. It
            may also be ``TChannel.FALLBACK`` if it's intended to be a
            catch-all endpoint.
        :param scheme:
            Name of the scheme under which the endpoint is being registered.
            One of "raw", "json", and "thrift". Defaults to "raw", except if
            "endpoint" was a module, in which case this defaults to "thrift".

        :param handler:
            If specified, this is the handler function. If ignored, this
            function returns a decorator that can be used to register the
            handler function.

        :returns:
            If ``handler`` was specified, this returns ``handler``. Otherwise,
            it returns a decorator that can be applied to a function to
            register it as the handler.
        """
        assert endpoint is not None, "endpoint is required"

        if endpoint is TChannel.FALLBACK:
            decorator = partial(self._handler.register, TChannel.FALLBACK)
            if handler is not None:
                return decorator(handler)
            else:
                return decorator

        if not scheme:
            # scheme defaults to raw, unless the endpoint is a service module.
            if inspect.ismodule(endpoint):
                scheme = "thrift"
            else:
                scheme = "raw"
        scheme = scheme.lower()
        if scheme == 'thrift':
            decorator = partial(self._register_thrift, endpoint, **kwargs)
        else:
            decorator = partial(
                self._register_simple, endpoint, scheme, **kwargs
            )

        if handler is not None:
            return decorator(handler)
        else:
            return decorator


class TChannelServer(tornado.tcpserver.TCPServer):
    __slots__ = ('tchannel',)

    def __init__(self, tchannel):
        super(TChannelServer, self).__init__()
        self.tchannel = tchannel

    @tornado.gen.coroutine
    def handle_stream(self, stream, address):
        log.debug("New incoming connection from %s:%d" % address)

        conn = StreamConnection(
            connection=stream,
            tchannel=self.tchannel,
            direction=INCOMING,
        )

        yield conn.expect_handshake(headers={
            'host_port': self.tchannel.hostport,
            'process_name': self.tchannel.process_name,
            'tchannel_language': TCHANNEL_LANGUAGE,
            'tchannel_language_version': TCHANNEL_LANGUAGE_VERSION,
            'tchannel_version': TCHANNEL_VERSION,
        })

        log.debug(
            "Successfully completed handshake with %s:%s (%s)",
            conn.remote_host,
            conn.remote_host_port,
            conn.remote_process_name)

        self.tchannel.peers.get(
            "%s:%s" % (conn.remote_host,
                       conn.remote_host_port)
        ).register_incoming_conn(conn)

        yield conn.serve(handler=self._handle)

    def _handle(self, message, connection):
        self.tchannel.receive_call(message, connection)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import random


# TODO: cython
def _uniq_id():
    """Create a random 64-bit signed integer.

    :rtype: unsigned long
    """
    return random.getrandbits(64)


class Trace(object):
    """A Zipkin trace/span."""

    __slots__ = [
        'trace_id',
        'name',
        'span_id',
        'parent_span_id',
        'endpoint',
        '_tracers',
        'traceflags',
        'annotations'
    ]

    def __init__(
        self,
        name=None,
        trace_id=None,
        span_id=None,
        parent_span_id=0,
        endpoint=None,
        traceflags=0
    ):
        """
        :param name:
            Name of the current span.

        :param trace_id:
            Optional trace id used reconcile all spans associated with a
            request. If not provided a random id will be generated.
        :type trace_id: int or None

        :param span_id:
            Optional span id for this particular span. If not provided a random
            id will be generatd.
        :type span_id: int or None

        :param parent_span_id:
            Optional parent span id for this particular span. It not provided
            will remain ``None``.
        :type parent_span_id: int or None

        :param endpoint:
            A default :py:class:`Endpoint` instance to associate with all
            annotations for this trace. If an :py:class:`Annotation` explicitly
            defines its own :py:class:`Endpoint`, it will override this value.
        :type endpoint: :py:class:`Endpoint`

        :param tracers:
            Tracers to record this Trace's events. Primarily useful for unit
            testing.
        :type traces: list or :py:class:`Tracer`
        """
        # If no trace_id and span_id are given we want to generate new
        # 64-bit integer ids.
        self.name = name
        self.trace_id = trace_id or _uniq_id()
        self.span_id = span_id or _uniq_id()

        # If no parent_span_id is given then we assume there is no parent span
        # and leave it as None.
        self.parent_span_id = parent_span_id or 0

        self.traceflags = traceflags
        self.endpoint = endpoint
        self.annotations = []

    def __eq__(self, other):
        return (
            (self.trace_id, self.span_id, self.parent_span_id) ==
            (other.trace_id, other.span_id, other.parent_span_id)
        )

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return (
            '{0.__class__.__name__}(name={0.name!r}, trace_id={0.trace_id!r}, '
            'span_id={0.span_id!r}, parent_span_id={0.parent_span_id!r})'
        ).format(self)

    def child(self, name, endpoint=None):
        """
        Create a new instance of this class derived from the current instance
        such that::

            new.trace_id == current.trace_id

        and::

            new.parent_span_id == current.span_id

        The new :py:class:`Trace` instance will have a new unique span_id and
        if set the endpoint of the current :py:class:`Trace` object.

        :rtype: Trace
        """
        trace = self.__class__(
            name,
            trace_id=self.trace_id,
            parent_span_id=self.span_id,
            endpoint=self.endpoint
        )

        return trace



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import tornado
import tornado.gen

from ..enum import enum
from ..errors import TChannelError
from ..messages.common import ChecksumType
from ..messages.common import FlagsType
from ..messages.common import StreamState
from ..serializer.raw import RawSerializer
from .stream import InMemStream
from .util import get_arg

StatusCode = enum(
    'StatusCode',
    ok=0x00,
    error=0x01,
)


class Response(object):
    """An outgoing response.

    Response class is used to represent the CallResponseMessage at User's
    level. This is going to hide the protocol level message information.
    """

    # TODO decide which elements inside "message" object to expose to user.
    def __init__(
        self,
        connection=None,
        id=None,
        flags=None,
        code=None,
        tracing=None,
        headers=None,
        checksum=None,
        argstreams=None,
        serializer=None,
    ):

        self.flags = flags or FlagsType.none
        self.code = code or StatusCode.ok
        self.tracing = tracing
        self.checksum = checksum or (ChecksumType.crc32c, 0)
        # argstreams is a list of InMemStream/PipeStream objects
        self.argstreams = argstreams or [InMemStream(),
                                         InMemStream(),
                                         InMemStream()]
        self.headers = headers or {}
        self.id = id
        self.connection = connection
        self.state = StreamState.init
        self.flushed = False

        self.serializer = serializer or RawSerializer()

    @property
    def status_code(self):
        return self.code

    @status_code.setter
    def status_code(self, status):
        if status not in StatusCode:
            raise TChannelError("Invalid status code!")

        self.code = status

    @property
    def ok(self):
        return self.code == StatusCode.ok

    def get_header_s(self):
        """Get the raw stream of header.

        :return: the argstream of header
        """
        return self.argstreams[1]

    def get_body_s(self):
        """Get the raw stream of body.

        :return: the argstream of body
        """
        return self.argstreams[2]

    @tornado.gen.coroutine
    def get_header(self):
        """Get the header value from the response.

        :return: a future contains the deserialized value of header
        """
        raw_header = yield get_arg(self, 1)
        if not self.serializer:
            raise tornado.gen.Return(raw_header)
        else:
            header = self.serializer.deserialize_header(raw_header)
            raise tornado.gen.Return(header)

    @tornado.gen.coroutine
    def get_body(self):
        """Get the body value from the response.

        :return: a future contains the deserialized value of body
        """

        raw_body = yield get_arg(self, 2)
        if not self.serializer:
            raise tornado.gen.Return(raw_body)
        else:
            body = self.serializer.deserialize_body(raw_body)
            raise tornado.gen.Return(body)

    def set_body_s(self, stream):
        """Set customized body stream.

        Note: the body stream can only be changed before the stream
        is consumed.

        :param stream: InMemStream/PipeStream for body

        :except TChannelError:
            Raise TChannelError if the stream is being sent when you try
            to change the stream.
        """
        if self.argstreams[2].state == StreamState.init:
            self.argstreams[2] = stream
        else:
            raise TChannelError(
                "Unable to change the body since the streaming has started")

    def set_header_s(self, stream):
        """Set customized header stream.

        Note: the header stream can only be changed before the stream
        is consumed.

        :param stream: InMemStream/PipeStream for header

        :except TChannelError:
            Raise TChannelError if the stream is being sent when you try
            to change the stream.
        """

        if self.argstreams[1].state == StreamState.init:
            self.argstreams[1] = stream
        else:
            raise TChannelError(
                "Unable to change the header since the streaming has started")

    def write_header(self, chunk):
        """Write to header.

        Note: the header stream is only available to write before write body.

        :param chunk: content to write to header

        :except TChannelError:
            Raise TChannelError if the response's flush() has been called
        """

        if self.serializer:
            header = self.serializer.serialize_header(chunk)
        else:
            header = chunk

        if self.flushed:
            raise TChannelError("write operation invalid after flush call")

        if (self.argstreams[0].state != StreamState.completed and
                self.argstreams[0].auto_close):
            self.argstreams[0].close()

        return self.argstreams[1].write(header)

    def write_body(self, chunk):
        """Write to header.

        Note: whenever write_body is called, the header stream will be closed.
        write_header method is unavailable.

        :param chunk: content to write to body

        :except TChannelError:
            Raise TChannelError if the response's flush() has been called
        """

        if self.serializer:
            body = self.serializer.serialize_body(chunk)
        else:
            body = chunk

        if self.flushed:
            raise TChannelError("write operation invalid after flush call")

        if (self.argstreams[0].state != StreamState.completed and
                self.argstreams[0].auto_close):
            self.argstreams[0].close()
        if (self.argstreams[1].state != StreamState.completed and
                self.argstreams[1].auto_close):
            self.argstreams[1].close()

        return self.argstreams[2].write(body)

    def flush(self):
        """Flush the response buffer.

        No more write or set operations is allowed after flush call.
        """
        self.flushed = True
        self.close_argstreams()

    def set_exception(self, exception, exc_info=None):
        for stream in self.argstreams:
            stream.set_exception(exception, exc_info=exc_info)
            stream.close()

    def close_argstreams(self, force=False):
        for stream in self.argstreams:
            if stream.auto_close or force:
                stream.close()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import struct

from .errors import ReadError

skip = '_'


def none():
    """A ReadWriter that consumes nothing and returns None."""
    return NoneReadWriter()


def constant(rw, value):
    """A ReadWriter that runs the given ReadWriter and ignores the value.

    Always writes and returns ``value`` instead.

    :param rw:
        ReadWriter to run
    :param value:
        Value to serialize and return
    """
    return ConstantReadWriter(rw, value)


def number(width_bytes):
    """Build a ReadWriter for integers of the given width.

    :param width_bytes:
        Width of the integer. One of 1, 2, 4 and 8.
    """
    return NumberReadWriter(width_bytes)


def args(length_rw):
    """Build a ReadWriter for args=[arg1, arg2, arg3]

    :param length_rw:
        ReadWriter for the length of each arg
    """
    return ArgsReaderWriter(length_rw)


def len_prefixed_string(length_rw, is_binary=False):
    """Build a ReadWriter for strings prefixed with their length.

    .. code-block:: python

        len_prefixed_string(number(2))  # == str~2

    :param length_rw:
        ReadWriter for the length of the string
    :param is_binary:
        Whether the string is a binary blob. If this is False (the default),
        the string will be encoded/decoded to UTF-8 before writing/reading.
    """
    return LengthPrefixedBlobReadWriter(length_rw, is_binary)


def chain(*rws):
    """Build a ReadWriter from the given list of ReadWriters.

    .. code-block:: python

        chain(
            number(1),
            number(8),
            len_prefixed_string(number(2)),
        )  # == n1:1 n2:8 s~2

    Reads/writes from the given ReadWriters in-order. Returns lists of values
    in the same order as the ReadWriters.

    :param rws:
        One or more ReadWriters
    """
    assert rws is not None
    if len(rws) == 1 and isinstance(rws[0], list):
        # In case someone does chain([l0, l1, ...])
        rws = rws[0]
    return ChainReadWriter(rws)


def dictionary(*pairs):
    """Build a ReadWriter that reads/writes dictionaries.

    ``pairs`` are tuples containing field names and their corresponding
    ReadWriters. The fields will be read and written in the same order
    provided here.

    For example the following ReadWriter will read and write dictionaries in
    the form ``{"flags": <byte>, "id": <int32>}``.

    .. code-block:: python

        dictionary(
            ("flags", number(1)),
            ("id", number(4)),
        )

    For pairs where the key name is `rw.skip`, the value will not be saved and
    the serializer will receive None.

    :param pairs:
        One or more tuples in the from ``(<field name>, <ReadWriter>)``.
    """
    return NamedChainReadWriter(pairs)


def instance(cls, *pairs):
    """Build a ReadWriter that reads/writes intances of the given class.

    ``pairs`` are key-value pairs that specify constructor argument names and
    their corresponding ReadWriters. These same names are used to access
    attributes on instances when writing.


    .. code-block:: python

        instance(
            Person,
            ("name", len_prefixed_string(number(2))),
            ("age", number(1))
        )

    For pairs where the attribute name is `rw.skip`, the value will not be
    passed to the constructor. Further, while serializing, None will be passed
    to the serializer.

    :param cls:
        A class with an ``__init__`` method accepting keyword arguments for
        all items specified in ``pairs``
    :param pairs:
        Key-value pairs mapping argument name to ReadWriter.
    """
    return InstanceReadWriter(cls, pairs)


def headers(length_rw, key_rw, value_rw=None):
    """Build a ReadWriter for header lists.

    A header is represented as::

        count:L (key:K value:V){count}

    The value produced is a list of key-value pairs. For example,

    .. code-block:: python

        headers(
            number(L),
            len_prefixed_string(number(K)),
            len_prefixed_string(number(V)),
        )

    :param length_rw:
        ReadWriter for the number of pairs in the header
    :param key_rw:
        ReadWriter for a key in a pair
    :param value_rw:
        ReadWriter for a value in a pair. Defaults to ``key_rw``.
    """
    return HeadersReadWriter(length_rw, key_rw, value_rw)


def switch(switch_rw, cases):
    """A ReadWriter that picks behavior based on the value of ``switch_rw``.

    .. code-block:: python

        switch(
            number(1), {
                0: option_1_rw(),
                1: option_2_rw()
            }
        )

    Produces a tuple in the from ``(switch_value, case_value)``. If a given
    switch value did not have a corresponding case, nothing will be written to
    the stream and None will be returned as the value when reading.

    :param switch_rw:
        A ReadWriter that produces a value to dispatch on
    :param cases:
        Pairs where the key is the expected value from ``switch_rw``. If the
        value matches, the corresponding ReadWriter will be executed.
    """
    return SwitchReadWriter(switch_rw, cases)


class ReadWriter(object):
    """Provides the ability to read/write types from/to file-like objects.

    ReadWrites SHOULD not maintain any state between calls to
    ``read``/``write`` and MUST be re-usable and thread-safe. The
    ``read``/``write`` methods MAY Be called on the same ReadWriter instance
    multiple times for different requests at the same time.

    The file-like stream object MUST provide ``read(int)`` and ``write(str)``
    methods with behaviors as follows:

    ``read(int)``
        MUST return the specified number of bytes from the stream. MAY return
        fewer bytes if the end of the stream was reached.
    ``write(str)``
        MUST write the given string or buffer to the stream.
    """

    def read(self, stream):
        """Read and return the object from the stream.

        :param stream:
            file-like object providing a `read(int)` method
        :returns: the deserialized object
        :raises ReadError:
            for parse errors or if the input is too short
        """
        raise NotImplementedError()

    def write(self, obj, stream):
        """Write the object to the stream.

        :param stream:
            file-like obect providing a `write(str)` method
        :returns:
            the stream
        """
        raise NotImplementedError()

    def length(self, obj):
        """Return the number of bytes will actually be written into io.

        For cases where the width depends on the input, this should return the
        length of data will be written into iostream."""
        raise NotImplementedError()

    def width(self):
        """Return the number of bytes this ReadWriter is expected to take.

        For cases where the width depends on the input, this should return the
        minimum width the ReadWriter is expected to take."""
        raise NotImplementedError()

    def take(self, stream, num):
        """Read the given number of bytes from the stream.

        :param stream:
            stream to read from
        :param num:
            number of bytes to read
        :raises ReadError:
            if the stream did not yield the exact number of bytes expected
        """
        s = stream.read(num)
        slen = len(s)
        if slen != num:
            raise ReadError(
                "Expected %d bytes but got %d bytes." % (num, slen)
            )
        return s


class DelegatingReadWriter(ReadWriter):
    """Allows mapping ReadWriters onto different types.

    A common pattern is to define a base ReadWriter using the primitives from
    this module and then map those onto custom types.

    For example, consider a Person class.

    .. code-block:: python

        Person = namedtuple('Person', 'name age')

    Given a ReadWriter that produces a ``(name, age)`` tuple, we want to map
    it to/from Person object.

    .. code-block:: python

        class PersonReadWriter(DelegatingReadWriter):
            __rw__ = # a ReadWriter that produces (name, age) tuples

            def read(self, stream):
                (name, age) = super(PersonReadWriter, self).read(stream)
                return Person(name, age)

            def write(self, person, stream):
                super(PersonReadWriter, self).write(
                    (person.name, person.age),
                    stream,
                )
    """

    # The underlying ReadWriter. All calls will be delegated to this.
    __rw__ = None

    class __metaclass__(type):

        def __new__(mcs, name, bases, dct):
            if bases != (ReadWriter,):
                # Children of this class MUST provide __rw__
                assert dct.get('__rw__'), (
                    "%s.__rw__ must be set" % name
                )
            return type.__new__(mcs, name, bases, dct)

    def read(self, stream):
        return self.__rw__.read(stream)

    def write(self, obj, stream):
        self.__rw__.write(obj, stream)
        return stream

    def width(self):
        return self.__rw__.width()

    def length(self, obj):
        return self.__rw__.length(obj)


class NumberReadWriter(ReadWriter):
    """See :py:func:`number` for documentation."""

    _FORMATS = {
        1: '>B',
        2: '>H',
        4: '>I',
        8: '>Q',
    }

    __slots__ = ('_width', '_format')

    def __init__(self, width_bytes):
        assert width_bytes in self._FORMATS, (
            "Unsupported integer width '%d'" % width_bytes
        )
        self._width = width_bytes
        self._format = self._FORMATS[width_bytes]

    def read(self, stream):
        return struct.unpack(self._format, self.take(stream, self._width))[0]

    def write(self, num, stream):
        # Cast to int just in case the value is still a float
        stream.write(struct.pack(self._format, int(num)))
        return stream

    def width(self):
        return self._width

    def length(self, obj):
        return self._width


class ArgsReaderWriter(ReadWriter):
    def __init__(self, length_rw, num=3):
        assert length_rw is not None
        self._length_rw = length_rw
        self._rw = len_prefixed_string(self._length_rw,
                                       is_binary=True)
        self.num = num

    def read(self, stream):
        args = []
        try:
            for _ in range(self.num):
                args.append(self._rw.read(stream))
        except ReadError:
            pass
        return args

    def write(self, args, stream):
        for arg in args:
            if arg is None:
                arg = ""
            self._rw.write(arg, stream)

    def width(self):
        return self.num * self._length_rw.width()

    def length(self, args):
        size = 0
        for arg in args:
            if arg is None:
                arg = ""
            size += self._rw.length(arg)

        return size


class LengthPrefixedBlobReadWriter(ReadWriter):
    """See :py:func:`len_prefixed_string` for documentation."""

    __slots__ = ('_length', '_is_binary')

    def __init__(self, length_rw, is_binary=False):
        assert length_rw is not None
        self._length = length_rw
        self._is_binary = is_binary

    def read(self, stream):
        length = self._length.read(stream)
        if length == 0:
            return ""
        else:
            blob = self.take(stream, length)
            if not self._is_binary:
                blob = blob.decode('utf-8')
            return blob

    def write(self, s, stream):
        if not self._is_binary:
            s = s.encode('utf-8')
        length = len(s)
        self._length.write(length, stream)
        stream.write(s)
        return stream

    def width(self):
        return self._length.width()

    def length(self, s):
        if not self._is_binary:
            s = s.encode('utf-8')

        return len(s) + self._length.width()


class ChainReadWriter(ReadWriter):
    """See :py:func:`chain` for documentation."""

    __slots__ = ('_links',)

    def __init__(self, links):
        assert links is not None
        self._links = tuple(links)

    def read(self, stream):
        return [link.read(stream) for link in self._links]

    def write(self, items, stream):
        assert len(items) == len(self._links)

        for item, link in zip(items, self._links):
            link.write(item, stream)
        return stream

    def width(self):
        return sum(link.width() for link in self._links)

    def length(self, items):
        assert len(items) == len(self._links)

        size = 0
        for item, link in zip(items, self._links):
            size += link.length(item)

        return size


class NamedChainReadWriter(ReadWriter):
    """See :py:func:`dictionary` for documentation."""

    __slots__ = ('_pairs',)

    def __init__(self, pairs):
        assert pairs is not None
        self._pairs = pairs

    def read(self, stream):
        result = {}
        for name, rw in self._pairs:
            try:
                value = rw.read(stream)
                if name != skip:
                    result[name] = value
            except ReadError as e:
                raise ReadError("Failed to read %s: %s" % (name, e))
        return result

    def write(self, obj, stream):
        for name, rw in self._pairs:
            if name != skip:
                rw.write(obj[name], stream)
            else:
                rw.write(None, stream)
        return stream

    def width(self):
        return sum(rw.width() for _, rw in self._pairs)

    def length(self, obj):
        size = 0
        for name, rw in self._pairs:
            if name != skip:
                size += rw.length(obj[name])
            else:
                size += rw.length(None)
        return size


class InstanceReadWriter(ReadWriter):

    __slots__ = ('_cls', '_pairs',)

    def __init__(self, cls, pairs):
        self._pairs = pairs
        self._cls = cls

    def read(self, stream):
        kwargs = {}
        try:
            for attr, rw in self._pairs:
                value = rw.read(stream)
                if attr != skip:
                    kwargs[attr] = value
        except ReadError as e:
            raise ReadError("Failed to read %s: %s" % (self._cls, e))

        return self._cls(**kwargs)

    def write(self, obj, stream):
        for attr, rw in self._pairs:
            if attr != skip:
                value = getattr(obj, attr)
                rw.write(value, stream)
            else:
                rw.write(None, stream)
        return stream

    def width(self):
        return sum(rw.width() for _, rw in self._pairs)

    def length(self, obj):
        size = 0
        for attr, rw in self._pairs:
            if attr != skip:
                value = getattr(obj, attr)
                size += rw.length(value)
            else:
                size += rw.length(None)

        return size

    def length_no_args(self, obj):
        size = 0
        for attr, rw in self._pairs:
            if attr == "args":
                continue
            if attr != skip:
                value = getattr(obj, attr)
                size += rw.length(value)
            else:
                size += rw.length(None)

        return size


class HeadersReadWriter(ReadWriter):
    """See :py:func:`headers` for documentation."""

    __slots__ = ('_length', '_key', '_value')

    def __init__(self, length_rw, key_rw, value_rw=None):
        self._length = length_rw
        self._pair = chain(key_rw, value_rw or key_rw)

    def read(self, stream):
        count = self._length.read(stream)
        headers = []
        for i in range(count):
            headers.append(self._pair.read(stream))
        return headers

    def write(self, headers, stream):
        # In case someone does write({..}, stream)
        if isinstance(headers, dict):
            headers = headers.items()

        self._length.write(len(headers), stream)
        for pair in headers:
            self._pair.write(pair, stream)
        return stream

    def width(self):
        return self._length.width()

    def length(self, headers):
        size = 0
        if isinstance(headers, dict):
            headers = headers.items()

        size += self._length.length(len(headers))
        for pair in headers:
            size += self._pair.length(pair)

        return size


class NoneReadWriter(ReadWriter):
    def read(self, stream):
        return None

    def write(self, _, stream):
        return stream

    def width(self):
        return 0

    def length(self, obj):
        return 0


class ConstantReadWriter(ReadWriter):

    __slots__ = ('_rw', '_value')

    def __init__(self, rw, value):
        self._rw = rw
        self._value = value

    def read(self, stream):
        self._rw.read(stream)
        return self._value

    def write(self, out, stream):
        self._rw.write(self._value, stream)
        return stream

    def width(self):
        return self._rw.width()

    def length(self, obj):
        return self._rw.width()


class SwitchReadWriter(ReadWriter):

    __slots__ = ('_switch', '_cases')

    def __init__(self, switch_rw, cases_rw):
        self._switch = switch_rw
        self._cases = cases_rw

    def read(self, stream):
        k = self._switch.read(stream)

        if k in self._cases:
            v = self._cases[k].read(stream)
            return (k, v)
        else:
            return (k, None)

    def write(self, item, stream):
        k, v = item
        self._switch.write(k, stream)
        if v is not None and k in self._cases:
            self._cases[k].write(v, stream)
        return stream

    def width(self):
        return self._switch.width()

    def length(self, item):
        k, v = item
        size = 0
        size += self._switch.length(k)
        if v is not None and k in self._cases:
            size += self._cases[k].length(v)

        return size



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, division, print_function, unicode_literals
)

from . import schemes
from . import transport as t
from .status import OK

__all__ = ['Response']


class Response(object):
    """A TChannel response.

    This is sent by handlers and received by callers.

    :ivar body:
        The payload of this response. The type of this attribute depends on the
        scheme being used (e.g., JSON, Thrift, etc.).

    :ivar headers:
        A dictionary of application headers. This should be a mapping of
        strings to strings.

    :ivar transport:
        Protocol-level transport headers. These are used for routing over
        Hyperbahn.
    """

    # TODO implement __repr__

    __slots__ = (
        'body',
        'status',
        'headers',
        'transport',
    )

    def __init__(self, body=None, headers=None, transport=None, status=None):
        if status is None:
            status = OK
        self.body = body
        self.status = status
        self.headers = headers
        self.transport = transport


class TransportHeaders(object):
    """Response-specific Transport Headers"""

    # TODO implement __repr__

    __slots__ = (
        'failure_domain',
        'scheme',
    )

    def __init__(self, failure_domain=None, scheme=None):
        if scheme is None:
            scheme = schemes.RAW

        self.failure_domain = failure_domain
        self.scheme = scheme

    @classmethod
    def from_dict(cls, data):
        return cls(
            failure_domain=data.get(t.FAILURE_DOMAIN),
            scheme=data.get(t.SCHEME),
        )

    def to_dict(self):
        m = {}

        if self.failure_domain is not None:
            m[t.FAILURE_DOMAIN] = self.failure_domain

        if self.scheme is not None:
            m[t.SCHEME] = self.scheme

        return m


def response_from_mixed(mixed):
    """Create Response from mixed input."""

    # if none then give empty Response
    if mixed is None:
        return Response()

    # if not Response, then treat like body
    if not isinstance(mixed, Response):
        return Response(mixed)

    # it's already a Response
    return mixed



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from collections import namedtuple

from . import rw
from .errors import ReadError
from .io import BytesIO

FrameHeader = namedtuple('FrameHeader', 'message_type message_id')
Frame = namedtuple('Frame', 'header payload')


class FrameReadWriter(rw.ReadWriter):

    # ReadWriter for Frame size
    size_rw = rw.number(2)  # size:2

    # ReadWriter for FrameHeaders
    header_rw = rw.instance(
        FrameHeader,
        ('message_type', rw.number(1)),             # type:1
        (rw.skip, rw.constant(rw.number(1), 0)),    # reserved:1
        ('message_id', rw.number(4)),               # id:4
        (rw.skip, rw.constant(rw.number(8), 0)),    # reserved:8
    )

    def read(self, stream, size=None):
        if not size:
            try:
                size = self.size_rw.read(stream)
            except ReadError:
                return None
        if not size:
            return None

        body = self.take(stream, size - self.size_rw.width())

        header_width = self.header_rw.width()
        header_body, payload = body[:header_width], body[header_width:]

        header = self.header_rw.read(BytesIO(header_body))
        return Frame(header, payload)

    def write(self, frame, stream):
        prelude_size = self.size_rw.width() + self.header_rw.width()
        size = prelude_size + len(frame.payload)

        self.size_rw.write(size, stream)
        self.header_rw.write(frame.header, stream)
        stream.write(frame.payload)

        return stream

    def width(self):
        return self.size_rw.width() + self.header_rw.width()

frame_rw = FrameReadWriter()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import logging

import opentracing
import tchannel.event
import tchannel.zipkin.trace
from opentracing.ext import tags
from opentracing.propagation import Format


class SpanWrapper(tchannel.zipkin.trace.Trace):
    """
    An extension of Trace class that captures OpenTracing Span object.
    Normally TChannel stores Trace object in the request context so that
    tracing information is propagated from server endpoint to client call
    sites. By storing an OpenTracing span in the same context we can
    integrate with non-TChannel RPC calls.
    """
    def __init__(self, span, *args, **kwargs):
        super(SpanWrapper, self).__init__(*args, **kwargs)
        self.span = span

    @staticmethod
    def get_span(tracing):
        return getattr(tracing, 'span', None)


# noinspection PyBroadException
class OpenTracingHook(tchannel.event.EventHook):
    """
    Generates server and client tracing spans using OpenTracing API.

    See `tests/test_opentracing.py` for example of integration with
    HTTP servers and clients using `opentracing_instrumentation` package.
    """

    def __init__(self, context_provider, tracer=None,
                 span_to_trace_fn=None,
                 log_exception_fn=None):
        """Manage OpenTracing spans

        :param context_provider:
            TChannel's RequestContextProvider
        :param tracer:
            An instance of OpenTracing Tracer
        :param span_to_trace_fn:
            An optional, Tracer-specific function that extracts trace_id,
            span_id, parent_span_id from the Span in order to pass
            those to the constructor of the TChannel's Trace class.
            The function must return a dictionary with keys trace_id,
            span_id, and parent_span_id, where values are all i64.
        :param log_exception_fn:
            Zero-arg function used to log exceptions. Defaults to
            logging.exception('OpenTracing exception').
        """
        self.context_provider = context_provider
        self._tracer = tracer
        self.span_to_trace_fn = span_to_trace_fn
        if not log_exception_fn:
            log_exception_fn = (
                lambda: logging.exception('OpenTracing exception'))
        self.log_exception_fn = log_exception_fn

    @staticmethod
    def _operation_name(request):
        if hasattr(request, 'endpoint'):
            return request.endpoint
        else:
            return 'tchannel::call'

    @property
    def tracer(self):
        return self._tracer if self._tracer else opentracing.tracer

    def _get_current_span(self):
        context = self.context_provider.get_current_context()
        if context is None:
            return None
        span = getattr(context, 'span', None)
        if span:
            return span
        parent_tracing = getattr(context, 'parent_tracing', None)
        if not parent_tracing:
            return None
        return getattr(parent_tracing, 'span', None)

    def _span_to_trace(self, span):
        trace_kwargs = {}
        try:
            if self.span_to_trace_fn:
                ids = self.span_to_trace_fn(span)
                trace_kwargs = {
                    'trace_id': long(ids.get('trace_id', 1)),
                    'span_id': long(ids.get('span_id', 2)),
                    'parent_span_id': long(ids.get('parent_span_id', 1) or 1),
                }
        except:
            self.log_exception_fn()
        return trace_kwargs

    def before_send_request(self, request):
        try:
            parent_span = self._get_current_span()
            span = self.tracer.start_span(
                operation_name=self._operation_name(request),
                parent=parent_span
            )
            span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_CLIENT)
            if hasattr(request, 'service'):
                span.set_tag(tags.PEER_SERVICE, request.service)

            trace_kwargs = self._span_to_trace(span=span)
            request.tracing = SpanWrapper(
                span,
                trace_id=trace_kwargs.get('trace_id'),
                span_id=trace_kwargs.get('span_id'),
                parent_span_id=trace_kwargs.get('parent_span_id'),
                traceflags=0
            )

            # inject span into headers
            carrier = {}
            self.tracer.inject(span, Format.TEXT_MAP, carrier)
            for k, v in carrier.iteritems():
                request.headers[k] = v
        except:
            self.log_exception_fn()

    def after_receive_response(self, request, response):
        try:
            span = SpanWrapper.get_span(request.tracing)
            if span:
                span.finish()
        except:
            self.log_exception_fn()

    def after_receive_error(self, request, error):
        try:
            span = SpanWrapper.get_span(request.tracing)
            if span:
                span.set_tag('error', 'true')
                span.set_tag('exception', error)
                span.finish()
        except:
            self.log_exception_fn()

    def before_receive_request(self, request):
        try:
            span = None
            try:
                span = self.tracer.join(
                    operation_name=self._operation_name(request),
                    format=Format.TEXT_MAP,
                    carrier=request.headers
                )
            except opentracing.UnsupportedFormatException:
                self.log_exception_fn()
                return  # tracer does not support basic text format
            except opentracing.InvalidCarrierException:
                self.log_exception_fn()
                return  # tracer does not like headers
            except opentracing.TraceCorruptedException:
                self.log_exception_fn()
                pass  # we can still start a new span
            except:
                self.log_exception_fn()

            if span is None:
                span = self.tracer.start_span(
                    operation_name=self._operation_name(request),
                )

            if span:
                span.set_tag(tags.SPAN_KIND, tags.SPAN_KIND_RPC_SERVER)
                caller_name = request.headers.get('cn')
                if caller_name:
                    span.set_tag(tags.PEER_SERVICE, caller_name)

                request.tracing = SpanWrapper(span)
        except:
            self.log_exception_fn()

    def after_send_response(self, response):
        try:
            span = SpanWrapper.get_span(response.tracing)
            if span:
                span.finish()
        except:
            self.log_exception_fn()

    def after_send_error(self, error):
        try:
            span = SpanWrapper.get_span(error.tracing)
            if span:
                span.set_tag('error', 'true')
                span.set_tag('exception', error)
                span.finish()
        except:
            self.log_exception_fn()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Copyright 2012 Rackspace Hosting, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import logging
import sys
from collections import defaultdict

from tchannel import retry
from tchannel.zipkin.annotation import Endpoint
from . import glossary
from .tcollector import TCollector
from .formatters import json_formatter
from .formatters import thrift_formatter
from .formatters import i64_to_base64

log = logging.getLogger('zipkin_tracing')

zipkin_log = logging.getLogger('zipkin')


class EndAnnotationTracer(object):
    """
    A tracer which collects all annotations for a trace until one of several
    possible "end annotations" is seen. An end annotation indicates that from
    the perspective of this tracer the trace is complete.
    """

    # Default list of end annotations.
    DEFAULT_END_ANNOTATIONS = (glossary.CLIENT_RECV, glossary.SERVER_SEND)

    def __init__(self, tracer, end_annotations=None):
        """
        :param tracer:
            An :py:class:`Tracer` to delegate to once an end annotation is
            seen.
        :param end_annotations:
            Names of possible end annotations. Defaults to
            ``DEFAULT_END_ANNOTATIONS``.
        :type end_annotations: list of strings
        """
        self._tracer = tracer
        self._end_annotations = end_annotations or self.DEFAULT_END_ANNOTATIONS
        self._annotations_for_trace = defaultdict(list)

    def record(self, traces):
        for (trace, annotations) in traces:
            trace_key = (trace.trace_id, trace.span_id)
            self._annotations_for_trace[trace_key].extend(annotations)

            for annotation in annotations:
                if annotation.name in self._end_annotations:
                    saved_annotations = self._annotations_for_trace[trace_key]

                    del self._annotations_for_trace[trace_key]

                    self._tracer.record([(trace, saved_annotations)])

                    break

            zipkin_log.debug(
                "%s: Sending trace: %s w/ %s",
                self.__class__.__name__,
                trace_key,
                annotations,
            )

    def flush(self):
        self._tracer.flush()


class RawZipkinTracer(object):
    """
    Send annotations to Zipkin as Base64-encoded Thrift via Python logging.

    This implementation logs all annotations immediately and does not implement
    buffering of any sort.
    """

    def __init__(self, logger):
        """
        :param logger:
            An :py:class:`logging.Logger` instance. A handler should be defined
            for this logger.
        """
        self._logger = logger

    def record(self, traces):
        for (trace, annotations) in traces:
            self._logger.info(
                thrift_formatter(trace, annotations, isbased64=True)
            )


class TChannelZipkinTracer(object):
    """
    Send annotations to Zipkin as Base64-encoded Thrift via TChannel.

    This implementation sends all annotations immediately and doesn't implement
    buffering of any sort.
    """

    def __init__(self, tchannel):
        """
        :param tchannel:
            A tchannel instance to send the trace info to zipkin server
        """

        self._tchannel = tchannel
        self.span_host = None

    def parse_host_port(self):
        ip, _, port = self._tchannel.hostport.rpartition(':')
        return Endpoint(ip, int(port), self._tchannel.name)

    def record(self, traces):

        def submit_callback(f):
            if f.exception():
                log.info(
                    'Failed to submit zipkin trace',
                    exc_info=f.exc_info()
                )
        if not self.span_host:
            self.span_host = self.parse_host_port()

        fus = []
        for (trace, annotations) in traces:
            f = self._tchannel.thrift(
                TCollector.submit(
                    thrift_formatter(
                        trace=trace,
                        annotations=annotations,
                        isbased64=False,
                        span_host=self.span_host,
                    )
                ),
                shard_key=i64_to_base64(trace.trace_id),
                retry_on=retry.NEVER,
                retry_limit=0,
                trace=False,
            )
            f.add_done_callback(submit_callback)
            fus.append(f)

        return fus


class ZipkinTracer(object):
    """
    Send annotations to Zipkin as Base64-encoded Thrift via Python logging.

    This is equivalent to

    .. code-block: python
        EndAnnotationTracer(
            BufferingTracer(
                RawZipkinTracer(logger)
            )
        )

    This implementation mostly exists for convenience.
    """

    def __init__(
        self,
        logger,
        end_annotations=None,
        max_traces=50,
    ):
        """
        :param logger: See :py:class:`RawZipkinTracer`.

        :param end_annotations: See :py:class:`EndAnnotationTracer`.

        :param max_traces: See :py:class:`BufferingTracer`.
        """
        self._tracer = EndAnnotationTracer(
            BufferingTracer(
                RawZipkinTracer(logger),
                max_traces=max_traces,
            ),
            end_annotations=end_annotations
        )

    def record(self, traces):
        return self._tracer.record(traces)

    def flush(self):
        self._tracer.flush()


class DebugTracer(object):
    """
    Send annotations immediately to a file-like destination in JSON format.

    All traces will be written immediately to the destination.
    """

    def __init__(self, destination=None):
        """
        :param destination:
            A file-like object to write JSON formatted traces to.
        """
        self.destination = destination or sys.stdout

    def record(self, traces):
        self.destination.write(json_formatter(traces))
        self.destination.write('\n')
        self.destination.flush()


class BufferingTracer(object):
    """
    Buffer traces and defer recording until `max_traces` have been received
    since the last trace was recorded.

    When `max_traces` is exceeded, all buffered traces will be flushed.  This
    means that for a max_traces of 5 if 10 traces are received, all 10 traces
    will be flushed to the next tracer.
    """

    def __init__(self, tracer, max_traces=50):
        """
        :param tracer:
            A :py:class:`Tracer` to record bufferred traces to.

        :param max_traces:
            The number of traces to buffer before recording occurs. Default 50.
        :type max_traces: int
        """
        self._max_traces = max_traces

        self._tracer = tracer
        self._buffer = []

    def flush(self):
        flushable = self._buffer
        self._buffer = []

        if flushable:
            self._tracer.record(flushable)

    def record(self, traces):
        self._buffer.extend(traces)

        if len(self._buffer) >= self._max_traces:
            self.flush()


_globalTracers = []


def set_tracers(tracers):
    global _globalTracers
    _globalTracers = tracers


def push_tracer(tracer):
    global _globalTracers
    _globalTracers.append(tracer)


def flush_tracers():
    for tracer in _globalTracers:
        try:
            tracer.flush()
        except AttributeError:
            pass


def get_tracers():
    return _globalTracers



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import fcntl
import socket
import struct


# TODO This module is unix-only. Figure out something for Windows.


def interface_ip(interface):
    """Determine the IP assigned to us by the given network interface."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    return socket.inet_ntoa(
        fcntl.ioctl(
            sock.fileno(), 0x8915, struct.pack('256s', interface[:15])
        )[20:24]
    )
    # Explanation:
    # http://stackoverflow.com/questions/11735821/python-get-localhost-ip
    # http://stackoverflow.com/questions/24196932/how-can-i-get-the-ip-address-of-eth0-in-python


def local_ip():
    """Get the local network IP of this machine"""
    ip = socket.gethostbyname(socket.gethostname())
    if ip.startswith('127.'):
        # Check eth0, eth1, eth2, en0, ...
        interfaces = [
            i + str(n) for i in ("eth", "en", "wlan") for n in xrange(3)
        ]  # :(
        for interface in interfaces:
            try:
                ip = interface_ip(interface)
                break
            except IOError:
                pass
    return ip



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division

import math
import six
from collections import deque


class NoMatchError(Exception):
    pass


class HeapOperation(object):
    """HeapOperation defines the interface on how to manipulate the heap.
    Any extension of heap class should implement all the heap operations in
    order to have complete heap functions.
    """

    def lt(self, i, j):
        """Compare the items in position i and j of the heap.

        :param i: the first item's position of the heap list
        :param j: the second item's position of the heap list
        :return true if items[i] < items[j], otherwise false.
        """
        raise NotImplementedError()

    def push(self, x):
        """Push a new item into heap"""
        raise NotImplementedError()

    def pop(self):
        """Pop an item from the heap"""
        raise NotImplementedError()

    def peek(self, i):
        """Peek at the item at the given position without removing it from the
        heap.

        :param i:
            0-indexed position of the iteam in the heap
        """
        raise NotImplementedError()

    def swap(self, i, j):
        """swap items between position i and j of the heap"""
        raise NotImplementedError()

    def size(self):
        """Return length of the heap."""
        raise NotImplementedError()


def init(h):
    """Initialize existing object into the heap."""
    # heapify
    n = h.size()
    for i in six.moves.range(int(math.floor(n/2)) - 1, -1, -1):
        down(h, i, n)


def push(h, x):
    """Push a new value into heap."""
    h.push(x)
    up(h, h.size()-1)


def pop(h):
    """Pop the heap value from the heap."""
    n = h.size() - 1
    h.swap(0, n)
    down(h, 0, n)
    return h.pop()


def remove(h, i):
    """Remove the item at position i of the heap."""
    n = h.size() - 1
    if n != i:
        h.swap(i, n)
        down(h, i, n)
        up(h, i)

    return h.pop()


def fix(h, i):
    """Rearrange the heap after the item at position i got updated."""
    down(h, i, h.size())
    up(h, i)


def up(h, child):
    while child > 0:
        parent = int(math.floor((child - 1) / 2))
        if not h.lt(child, parent):
            break

        h.swap(parent, child)
        child = parent


def down(h, parent, n):
    while True:
        child1 = 2 * parent + 1
        if child1 >= n or child1 < 0:
            break

        min_child = child1
        child2 = child1 + 1
        if child2 < n and not h.lt(child1, child2):
            min_child = child2

        if not h.lt(min_child, parent):
            break

        h.swap(parent, min_child)
        parent = min_child


def smallest(heap, predicate):
    """Finds the index of the smallest item in the heap that matches the given
    predicate.

    :param heap:
        Heap on which this search is being performed.
    :param predicate:
        Function that accepts an item from the heap and returns true or false.
    :returns:
        Index of the first item for which ``predicate`` returned true.
    :raises NoMatchError:
        If no matching items were found.
    """
    n = heap.size()

    # items contains indexes of items yet to be checked.
    items = deque([0])
    while items:
        current = items.popleft()
        if current >= n:
            continue

        if predicate(heap.peek(current)):
            return current

        child1 = 2 * current + 1
        child2 = child1 + 1

        if child1 < n and child2 < n and heap.lt(child2, child1):
            # make sure we check the smaller child first.
            child1, child2 = child2, child1

        if child1 < n:
            items.append(child1)

        if child2 < n:
            items.append(child2)

    raise NoMatchError()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

from . import common
from .. import rw
from ..glossary import DEFAULT_TIMEOUT
from .call_request_continue import CallRequestContinueMessage
from .types import Types


class CallRequestMessage(CallRequestContinueMessage):
    """Initiate an RPC call."""
    message_type = Types.CALL_REQ

    __slots__ = CallRequestContinueMessage.__slots__ + (
        'ttl',
        'tracing',
        'service',
        'headers',
    )

    def __init__(
        self,
        flags=0,
        ttl=DEFAULT_TIMEOUT,
        tracing=None,
        service=None,
        headers=None,
        checksum=None,
        args=None,
        id=0,
    ):
        args = args or ["", "", ""]
        super(CallRequestMessage, self).__init__(flags, checksum, args, id)
        self.ttl = ttl
        self.tracing = tracing or common.Tracing(0, 0, 0, 0)
        self.service = service or ''
        self.headers = dict(headers) if headers else {}

call_req_rw = rw.instance(
    CallRequestMessage,
    ("flags", rw.number(1)),    # flags:1
    ("ttl", rw.number(4)),      # ttl:4
    ("tracing", common.tracing_rw),     # tracing:24
                                        # traceflags: 1
    ("service", rw.len_prefixed_string(rw.number(1))),  # service~1
    ("headers", rw.headers(             # nh:1 (hk~1 hv~1){nh}
        rw.number(1),
        rw.len_prefixed_string(rw.number(1))
    )),
    ("checksum", common.checksum_rw),   # csumtype:1 (csum:4){0, 1}
    ("args",
     rw.args(rw.number(2))),  # [arg1~2, arg2~2, arg3~2]
)



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import

import threading
from tornado.stack_context import StackContext


class TChannelLocal(threading.local):
    __slots__ = ('context',)

    def __init__(self):
        self.context = None

_LOCAL = TChannelLocal()


class RequestContext(object):
    """Tracks the :py:class:`Request` currently being handled.

    The asynchronous nature of Tornado means that multiple requests can be
    in-flight at any given moment. It's often useful to be able to see some
    information about the request that triggered the current method invocation.

    There are two ways to do this:

    * Pass the :py:class:`tchannel.Request` to every method that may need to
      use it.  This is performant but breaks MVC boundaries.

    * Use :py:class:`RequestContext` -- in particular
      :py:func:`get_current_context` -- to see this info from any point in your
      code. This can be "easier" (read: magical).

    :py:class:`RequestContext` uses Tornado's ``StackContext`` functionality,
    which hurts throughput. There's currently no way to disable
    :py:class:`RequestContext` tracking (for cases when you want to pass the
    :py:class:`tchannel.Request` explicity), although it is planned.


    :ivar parent_tracing:
        Tracing information (trace id, span id) for this request.
    """

    __slots__ = ('parent_tracing', '_old_context',)

    def __init__(self, parent_tracing=None):
        self.parent_tracing = parent_tracing
        self._old_context = None

    def __enter__(self):
        self._old_context = _LOCAL.context
        _LOCAL.context = self

    def __exit__(self, type, value, traceback):
        _LOCAL.context = self._old_context


# noinspection PyMethodMayBeStatic
class RequestContextProvider(object):
    def get_current_context(self):
        """
        :return: The current :py:class:`RequestContext` for this thread.
        """
        return _LOCAL.context

    def request_context(self, parent_tracing):
        """
        Factory method meant to be used as:

        .. code-block:: python

            with tchannel.context_provider.request_context(parent_tracing):
                handler_fn()

        :param parent_tracing:
        :return:
        """
        # TODO should this be using a thread-safe version of StackContext?
        return StackContext(lambda: RequestContext(parent_tracing))



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import (
    absolute_import, unicode_literals, division, print_function
)

import threading
from collections import namedtuple

from tornado.ioloop import IOLoop
from tornado.queues import QueueEmpty
from tornado.concurrent import Future

__all__ = ['Queue', 'QueueEmpty']


Node = namedtuple('Node', 'value next')


class Queue(object):
    """An unbounded, thread-safe asynchronous queue."""

    __slots__ = ('_get', '_put', '_lock')

    # How this works:
    #
    # _get and _put are futures maintaining pointers to a linked list of
    # futures. The linked list is implemented as Node objects holding the
    # value and the next future.
    #
    #     Node
    #   +---+---+   +---+---+  E: Empty future
    #   | 1 | F-|-->| 2 | E |  F: Filled future
    #   +---+---+   +---+---+
    #         ^           ^
    #   +---+ |     +---+ |
    #   | F-|-+     | F-|-+
    #   +---+       +---+
    #    _get        _put
    #
    # When there's a put, we fill the current empty future with a Node
    # containing the value and a pointer to the next, newly created empty
    # future.
    #
    #   +---+---+   +---+---+   +---+---+
    #   | 1 | F-|-->| 2 | F-|-->| 3 | E |
    #   +---+---+   +---+---+   +---+---+
    #         ^                       ^
    #   +---+ |                 +---+ |
    #   | F-|-+                 | F-|-+
    #   +---+                   +---+
    #    _get                    _put
    #
    # When there's a get, we read the value from the current Node, and move
    # _get to the next future.
    #
    #   +---+---+   +---+---+
    #   | 2 | F-|-->| 3 | E |
    #   +---+---+   +---+---+
    #         ^           ^
    #   +---+ |     +---+ |
    #   | F-|-+     | F-|-+
    #   +---+       +---+
    #    _get        _put

    def __init__(self):
        self._lock = threading.Lock()

        # Space for the next Node.
        hole = Future()

        # Pointer to the Future that will contain the next Node.
        self._get = Future()
        self._get.set_result(hole)

        # Pointer to the next empty Future that should be filled with a Node.
        self._put = Future()
        self._put.set_result(hole)

    def put(self, value):
        """Puts an item into the queue.

        Returns a Future that resolves to None once the value has been
        accepted by the queue.
        """
        io_loop = IOLoop.current()
        new_hole = Future()

        new_put = Future()
        new_put.set_result(new_hole)

        with self._lock:
            self._put, put = new_put, self._put

        answer = Future()

        def _on_put(future):
            if future.exception():  # pragma: no cover (never happens)
                return answer.set_exc_info(future.exc_info())

            old_hole = put.result()
            old_hole.set_result(Node(value, new_hole))
            answer.set_result(None)

        io_loop.add_future(put, _on_put)
        return answer

    def get_nowait(self):
        """Returns a value from the queue without waiting.

        Raises ``QueueEmpty`` if no values are available right now.
        """
        new_get = Future()

        with self._lock:
            if not self._get.done():
                raise QueueEmpty
            get, self._get = self._get, new_get

        hole = get.result()
        if not hole.done():
            # Restore the unfinished hole.
            new_get.set_result(hole)
            raise QueueEmpty

        value, new_hole = hole.result()
        new_get.set_result(new_hole)
        return value

    def get(self):
        """Gets the next item from the queue.

        Returns a Future that resolves to the next item once it is available.
        """
        io_loop = IOLoop.current()
        new_get = Future()

        with self._lock:
            get, self._get = self._get, new_get

        answer = Future()

        def _on_node(future):
            if future.exception():  # pragma: no cover (never happens)
                return answer.set_exc_info(future.exc_info())

            value, new_hole = future.result()
            new_get.set_result(new_hole)
            answer.set_result(value)

        def _on_get(future):
            if future.exception():  # pragma: no cover (never happens)
                return answer.set_exc_info(future.exc_info())

            hole = future.result()
            io_loop.add_future(hole, _on_node)

        io_loop.add_future(get, _on_get)
        return answer



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from tornado import gen

from . import THRIFT


class ThriftArgScheme(object):
    """Handler registration and serialization for Thrift.

    Use :py:func:`tchannel.thrift.load` to parse your Thrift IDL and compile
    it into a module dynamically.

    .. code:: python

        from tchannel import thrift

        keyvalue = thrift.load('keyvalue.thrift', service='keyvalue')

    To register a Thrift handler, use the ``register()`` decorator, providing
    a reference to the compiled service as an argument. The name of the
    service method should match the name of the decorated function.

    .. code:: python

        tchannel = TChannel(...)

        @tchannel.thrift.register(keyvalue.KeyValue)
        def setValue(request):
            data[request.body.key] = request.body.value

    Use methods on the compiled service to generate requests to remote
    services and execute them via ``TChannel.thrift()``.

    .. code:: python

        response = yield tchannel.thrift(
            keyvalue.KeyValue.setValue(key='foo', value='bar')
        )
    """

    NAME = THRIFT

    def __init__(self, tchannel):
        self._tchannel = tchannel

    @gen.coroutine
    def __call__(
        self,
        request,
        headers=None,
        timeout=None,
        retry_on=None,
        retry_limit=None,
        shard_key=None,
        trace=None,
        hostport=None,
        routing_delegate=None,
    ):
        """Make a Thrift TChannel request.

        Returns a ``Response`` containing the return value of the Thrift
        call (if any). If the remote server responded with a Thrift exception,
        that exception is raised.

        :param string request:
            Request obtained by calling a method on service objects generated
            by :py:func:`tchannel.thrift.load`.
        :param dict headers:
            Dictionary of header key-value pairs.
        :param float timeout:
            How long to wait (in seconds) before raising a ``TimeoutError`` -
            this defaults to ``tchannel.glossary.DEFAULT_TIMEOUT``.
        :param string hostport:
            A 'host:port' value to use when making a request directly to a
            TChannel service, bypassing Hyperbahn. This value takes precedence
            over the ``hostport`` specified to
            :py:func:`tchannel.thrift.load`.
        :param string retry_on:
            What events to retry on - valid values can be found in
            ``tchannel.retry``.
        :param int retry_limit:
            How many attempts should be made (in addition to the initial
            attempt) to re-send this request when retryable error conditions
            (specified by ``retry_on``) are encountered.

            Defaults to ``tchannel.retry.DEFAULT_RETRY_LIMIT`` (4).

            Note that the maximum possible time elapsed for a request is thus
            ``(retry_limit + 1) * timeout``.
        :param routing_delegate:
            Name of a service to which the request router should forward the
            request instead of the service specified in the call req.

        :rtype: Response
        """
        if not headers:
            headers = {}

        serializer = request.get_serializer()

        # serialize
        try:
            headers = serializer.serialize_header(headers=headers)
        except (AttributeError, TypeError):
            raise ValueError(
                'headers must be a map[string]string (a shallow dict'
                ' where keys and values are strings)'
            )

        body = serializer.serialize_body(request.call_args)

        # TODO There's only one yield. Drop in favor of future+callback.
        response = yield self._tchannel.call(
            scheme=self.NAME,
            service=request.service,
            arg1=request.endpoint,
            arg2=headers,
            arg3=body,
            timeout=timeout,
            retry_on=retry_on,
            retry_limit=retry_limit,
            hostport=hostport or request.hostport,
            shard_key=shard_key,
            trace=trace,
            routing_delegate=routing_delegate,
        )

        response.headers = serializer.deserialize_header(
            headers=response.headers
        )
        body = serializer.deserialize_body(body=response.body)

        response.body = request.read_body(body)
        raise gen.Return(response)

    def register(self, thrift_module, **kwargs):
        # dat circular import
        from tchannel.thrift import rw as thriftrw

        if isinstance(thrift_module, thriftrw.Service):
            # Dirty hack to support thriftrw and old API
            return thriftrw.register(
                # TODO drop deprecated tchannel
                self._tchannel._dep_tchannel._handler,
                thrift_module,
                **kwargs
            )
        else:
            return self._tchannel.register(
                scheme=self.NAME,
                endpoint=thrift_module,
                **kwargs
            )



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
VCR
===

``tchannel.testing.vcr`` provides VCR-like functionality for TChannel. Its
API is heavily inspired by the `vcrpy <https://github.com/kevin1024/vcrpy/>`_
library.

This allows recording TChannel requests and their responses into YAML files
during integration tests and replaying those recorded responses when the tests
are run next time.

The simplest way to use this is with the :py:func:`use_cassette` function.

.. autofunction:: use_cassette

Configuration
-------------

.. py:data:: tchannel.testing.vcr.DEFAULT_MATCHERS

    A tuple containing the default list of matchers used by
    :py:func:`tchannel.testing.vcr.use_cassette`.

Record Modes
~~~~~~~~~~~~

.. autoclass:: RecordMode

"""

from __future__ import absolute_import

from .config import use_cassette
from .record_modes import RecordMode
from .cassette import DEFAULT_MATCHERS


__all__ = ['use_cassette', 'RecordMode', 'DEFAULT_MATCHERS']
